from pathlib import Path
from datetime import datetime, timezone
from ..utilities import sxs_directory, sxs_identifier_re, path_to_invenio
from ..metadata import Metadata

def file_upload_allowed(file, directory_listing):
    """Return True if the file should be uploaded
    
    A file should be uploaded if
        * it is named "metadata.json" or "Horizons.h5"
        * it is named "Strain_*.json" or "ExtraWaveforms.json" and the corresponding
          ".h5" file is in the directory listing
        * it is named "Strain_*.h5" or "ExtraWaveforms.h5" and the corresponding
          ".json" file is in the directory listing
    
    """
    # Check `file.name` to ignore the directory
    if file.name in ["metadata.json", "Horizons.h5"]:
        return True
    if file.name.startswith("Strain_") or file.name.startswith("ExtraWaveforms"):
        # Ensure that both `.h5` and `.json` exist for all such files
        if file.suffix == ".json":
            return file.with_suffix(".h5") in directory_listing
        elif file.suffix == ".h5":
            return file.with_suffix(".json") in directory_listing
        else:
            return False
    return False


def files_to_upload(directory, annex_dir="."):
    """Return a list of files to upload

    The files to upload are those that are in the directory listing
    and pass the `file_upload_allowed` function.

    """
    full_directory = annex_dir / Path(directory)
    files = []
    for lev in full_directory.resolve().glob("Lev*"):
        directory_listing = list(lev.iterdir())
        files.extend([
            file for file in directory_listing
            if file_upload_allowed(file, directory_listing)
        ])
    return sorted(files, key=lambda x: str(x).lower())


def extract_id_from_common_metadata(file, annex_dir):
    """Extract the SXS ID from a common-metadata.txt file
    
    If the ID doesn't exist, return the directory path, relative to
    the `annex_dir`.
    """
    file = Path(file)
    annex_dir = Path(annex_dir)
    key = str(file.resolve().parent.relative_to(annex_dir.resolve()))
    with file.open("r") as f:
        for line in f.readlines():
            line = line.strip()
            if "alternative-names" in line:
                if (m := sxs_identifier_re.search(line)):
                    key = m["sxs_identifier"]
                    break
    return key


def local_simulations(annex_dir, compute_md5=False, show_progress=False):
    """
    Walk the annex directory to find and process all simulations

    For each `common-metadata.txt` file found:
        - Ensures that at least one directory starting with "Lev"
          exists; if not, the process is skipped.
        - Defines a key for the metadata, which is either:
            - The SXS ID contained in that file's "alternative-names"
              field, if present.
            - The directory path relative to `annex_dir`.
        - Chooses the highest "Lev" directory and extracts the
          metadata.
        - Finds all files to upload in the directory; if none are
          found, the process is skipped.
        - Adds the "files" dictionary to the metadata, pointing to
          each file that would be uploaded if the simulation were
          published.

    Parameters
    ----------
    annex_dir : (str or Path)
        The path to the annex directory to be processed.
    compute_md5 : bool, optional
        Whether to compute the MD5 hash of each file.  Default is
        False.
    show_progress : bool, optional
        Whether to show a progress bar.  Default is False.

    Returns
    -------
    dict :
        A dictionary containing the processed metadata.
    """
    from os import walk
    from ..utilities import md5checksum
    from tqdm import tqdm

    simulations = {}
    annex_dir = Path(annex_dir).resolve()

    if show_progress:  # Count the number of common-metadata.txt files
        num_files = 0
        for dirpath, dirnames, filenames in walk(annex_dir, topdown=True):
            if Path(dirpath).name.startswith("."):
                dirnames[:] = []
                continue
            if "common-metadata.txt" in filenames:
                if not any(d.startswith("Lev") for d in dirnames):
                    continue
                num_files += 1
                dirnames[:] = []
        progress_bar = tqdm(total=num_files, desc="Processing simulations")

    # The `walk` method can be made *much* faster than the `glob` method
    for dirpath, dirnames, filenames in walk(annex_dir, topdown=True):
        dirpath = Path(dirpath)

        # Ignore hidden directories
        if dirpath.name.startswith("."):
            dirnames[:] = []
            continue

        if "common-metadata.txt" in filenames:
            if not any(d.startswith("Lev") for d in dirnames):
                continue

            if show_progress:
                progress_bar.update(1)

            try:
                key = extract_id_from_common_metadata(dirpath / "common-metadata.txt", annex_dir)

                # Find the highest Lev directory and extract the metadata
                highest_lev = sorted(
                    [d for d in dirnames if d.startswith("Lev")]
                )[-1]
                metadata = Metadata.load(dirpath / highest_lev / "metadata")
                metadata = metadata.add_standard_parameters()

                metadata["directory"] = str(dirpath.relative_to(annex_dir))

                simulations[key] = metadata

                files = files_to_upload(dirpath, annex_dir)

                metadata["mtime"] = datetime.fromtimestamp(
                    max(
                        (
                            file.resolve().stat().st_mtime
                            for file in files
                            if file.exists()
                        ),
                        default=0.0,
                    ),
                    tz=timezone.utc,
                ).isoformat()

                metadata["files"] = {
                    path_to_invenio(file.relative_to(dirpath)): {
                        "link": str(file),
                        "size": file.stat().st_size,
                        "checksum": md5checksum(file) if compute_md5 else "",
                    }
                    for file in files
                    if file.exists()
                }
            except KeyboardInterrupt:
                raise
            except Exception as e:
                print(f"Error processing {dirpath}: {e}")

            dirnames[:] = []  # Don't keep looking for common-metadata.txt files under this directory

    return simulations


def write_local_simulations(annex_dir, output_file=None, compute_md5=False, show_progress=False):
    """Write the local simulations to a file for use when loading `Simulations`

    This function calls `local_simulations` to obtain the dictionary,
    but also writes the dictionary to a JSON file.

    Parameters
    ----------
    annex_dir : (str or Path)
        The path to the annex directory to be processed.
    output_file : (str or Path, optional)
        The path to the file to be written.  By default, the file is
        written to `sxs_directory("cache") / "local_simulations.json"`.
        N.B.: If you specify a different file, `sxs.load` will not
        automatically find it.
    compute_md5 : bool, optional
        Whether to compute the MD5 hash of each file.  Default is
        False.
    show_progress : bool, optional
        Whether to show a progress bar.  Default is False.

    Returns
    -------
    dict :
        A dictionary containing the processed metadata.
    """
    from json import dump

    # Process the annex directory to find all simulations
    simulations = local_simulations(annex_dir, compute_md5=compute_md5, show_progress=show_progress)

    # Write the simulations to file
    if output_file is not False:  # Test literal identity to allow `None`
        if output_file is None:
            output_file = sxs_directory("cache") / "local_simulations.json"
        else:
            output_file = Path(output_file)
        output_file.parent.mkdir(parents=True, exist_ok=True)
        with output_file.open("w") as f:
            dump(simulations, f, indent=2, separators=(",", ": "), ensure_ascii=True)

    return simulations



"""Simple regexes to understand SXS IDs"""

import re
import os

sep_regex = r"(:|/)" if os.sep == "/" else r"(:|/|\\)"

sxs_identifier_regex = (
    r"(?P<sxs_identifier>SXS:(?P<simulation_type>BBH|BHNS|NSNS)(?:_ExtCCE)?:"
    r"(?P<sxs_number>[0-9]+))(?:(v|V)(?P<version>[0-9.]+))?"
)
lev_regex = r"Lev(?P<lev>-?[0-9]+)"
sxs_id_version_lev_regex = sxs_identifier_regex + rf"(?:{sep_regex}{lev_regex})?"
sxs_id_version_lev_exact_regex = f"^{sxs_id_version_lev_regex}$"

file_regex = r"(?P<file>[a-zA-Z0-9_]+\.[a-zA-Z0-9]+)"
sxs_path_regex = sxs_id_version_lev_regex + rf"(?:{sep_regex}{file_regex})?"

sxs_identifier_re = re.compile(sxs_identifier_regex)
lev_re = re.compile(lev_regex)
lev_path_re = re.compile(f"{sep_regex}{lev_regex}")
sxs_id_version_lev_re = re.compile(sxs_id_version_lev_regex)
sxs_id_version_lev_exact_re = re.compile(sxs_id_version_lev_exact_regex)
sxs_path_re = re.compile(sxs_path_regex)

def sxs_id(s, default="", include_version=False):
    """Return the SXS ID contained in the input string

    An SXS ID is anything that matches the following regular expression:

        SXS:(BBH|BHNS|NSNS)(_ExtCCE)?:[0-9]+(v[0-9.]+)?

    If no match is found, the empty string is returned.  If multiple matches are
    found, only the first is returned.

    The input parameter may be:
    1) A string believed to contain the SXS ID
    2) The path to a file (like common-metadata.txt) containing such a string
    3) An object with a 'title' attribute
    4) An object with a 'title' item

    """
    id, version = sxs_id_and_version(s, default)
    if include_version:
        return f"{id}{version}"
    else:
        return id


def sxs_id_and_version(s, default=""):
    """Return the SXS ID and version contained in the input string"""
    import os.path
    import re
    try:
        s = s["title"]
    except (TypeError, KeyError):
        pass
    if hasattr(s, "title") and not isinstance(s.title, type("a".title)):
        s = s.title
    try:
        if os.path.isfile(s):
            with open(s, "r") as f:
                s = [l.strip() for l in f.splitlines()]
            for line in s:
                sxs_id_line = sxs_id_and_version(line)
                if sxs_id_line[0]:
                    return sxs_id_line
            return default, ""
    except TypeError:
        pass
    m = re.search(sxs_identifier_regex, s)
    if m:
        return m["sxs_identifier"], (f"v{m['version']}" if m["version"] else "")
    else:
        return default, ""


def simulation_title(sxs_id):
    """Create simulation title given just the SXS ID

    Currently supported types return output that looks like

        Binary black-hole simulation SXS:BBH:0001
        Black-hole neutron-star binary simulation SXS:BHNS:0001
        Binary neutron-star simulation SXS:NSNS:0001

    """
    import re
    sxs_system_re = re.compile(sxs_identifier_regex)
    m = sxs_system_re.search(sxs_id)
    if not m:
        raise ValueError(f"No SXS identifier found in '{sxs_id}'")
    sxs_identifier = m["sxs_identifier"]
    simulation_type = m["simulation_type"]
    if simulation_type == "BBH":
        title = f"Binary black-hole simulation {sxs_identifier}"
    elif simulation_type == "BHNS":
        title = f"Black-hole neutron-star binary simulation {sxs_identifier}"
    elif simulation_type == "NSNS":
        title = f"Binary neutron-star simulation {sxs_identifier}"
    else:
        raise ValueError(f"Did not recognize SXS system type '{simulation_type}'; should be BBH, BHNS, or NSNS.")
    return title


def lev_number(s):
    """Return the integer Lev number contained in the input string

    The Lev number is anything that matches the following regular expression:

        Lev(?P<lev>-?[0-9]+)

    If no match is found, `None` is returned.  If multiple matches are found,
    only the first is returned.

    """
    import re
    m = re.search(lev_regex, s)
    if m:
        return int(m["lev"])
    else:
        return None


def sxs_id_to_url(sxs_id):
    """Return the DOI URL for the SXS ID"""
    return f"https://doi.org/10.26138/{sxs_id}"



from ..utilities.string_converters import *
import numpy as np

class MetadataMetric:
    """A metric for comparing metadata.

    This class is designed to be used as a callable object that takes
    two collections of metadata (`sxs.Metadata`, `dict`, `pd.Series`)
    and returns a number measuring the distance between the metadata.
    
    With the default arguments, this will not strictly be a metric, as
    it does not satisfy the triangle inequality.  However, it is
    intended to be used as a heuristic for sorting and filtering
    metadata, rather than as a strict metric for clustering or
    classification.

    Note that calling an object of this class with two metadata
    collections will return the *squared* distance between them.

    Parameters
    ----------
    parameters : list of str, optional
        The names of the metadata fields to be compared.  The defaults
        are the reference quantities for mass ratio, spin,
        eccentricity, and mean anomaly.  Note that all of these fields
        *must* be present in *both* metadata collections.  (The
        `Metadata.add_standard_parameters` method may be useful here.)
    metric : array_like, optional
        The matrix used to weight the differences in the parameters.
        The default is a diagonal matrix with ones on the diagonal,
        except for the mean-anomaly entry, which is 1/pi^2.
    allow_different_object_types : bool, optional
        If True, metadata with different object types (BHBH, BHNS,
        NSNS) will be compared without penalty.  If False, metadata
        with different object types will be assigned an infinite
        distance.
    eccentricity_threshold1 : float, optional
        The threshold eccentricity below which we consider metadata1
        non-eccentric.  Default is 1e-2.
    eccentricity_threshold2 : float, optional
        The threshold eccentricity below which we consider metadata2
        non-eccentric.  Default is 1e-3.
    eccentricity_threshold_penalize_shorter : int, optional
        The number of orbits below which we penalize metadata2 for
        having a non-zero eccentricity when metadata1 does not.  This
        is intended to avoid ascribing small distances to systems with
        shorter inspirals.  Default is 20.

    The mean anomaly, if present, is treated specially to account for
    the fact that a mean anomaly of 0 is equivalent to a mean anomaly
    of 2π.  The difference between the entries in the two metadata
    collections is "unwrapped" before the metric is applied.

    If the eccentricity of metadata1 is below
    `eccentricity_threshold1`, then the mean anomaly is ignored.  If
    that is true and the eccentricity of metadata2 is below
    `eccentricity_threshold2` *and* the number of orbits in metadata2
    is longer than `eccentricity_threshold_penalize_shorter`, then the
    eccentricity is also ignored.  You may set these arguments to 0 to
    disable these features.

    """
    def __init__(
            self,
            parameters=[
                "reference_mass1",
                "reference_mass2",
                "reference_dimensionless_spin1",
                "reference_dimensionless_spin2",
                "reference_eccentricity",
                "reference_mean_anomaly",
            ],
            metric=np.diag([1, 1, 1, 1, 1, 1, 1, 1, 1, 1/np.pi**2]),
            allow_different_object_types=False,
            eccentricity_threshold1=1e-2,
            eccentricity_threshold2=1e-3,
            eccentricity_threshold_penalize_shorter=20,
    ):
        self.parameters = parameters
        self.metric = metric
        self.allow_different_object_types = allow_different_object_types
        self.eccentricity_threshold1 = eccentricity_threshold1
        self.eccentricity_threshold2 = eccentricity_threshold2
        self.eccentricity_threshold_penalize_shorter = eccentricity_threshold_penalize_shorter

    def __call__(self, metadata1, metadata2, debug=False):
        if not self.allow_different_object_types:
            type1 = (
                metadata1["object_types"]
                if "object_types" in metadata1
                else "".join(sorted([
                    metadata1.get("object1", "A").upper(),
                    metadata1.get("object2", "B").upper()
                ]))
            )
            type2 = (
                metadata2["object_types"]
                if "object_types" in metadata2
                else "".join(sorted([
                    metadata2.get("object1", "C").upper(),
                    metadata2.get("object2", "D").upper()
                ]))
            )
            if type1 != type2:
                return np.inf

        values1 = [metadata1[parameter] for parameter in self.parameters]
        values2 = [metadata2[parameter] for parameter in self.parameters]

        if debug:
            print(f"{self.parameters=}")
            print(f"{values1=}")
            print(f"{values2=}")

        if "reference_mean_anomaly" in self.parameters:
            i = self.parameters.index("reference_mean_anomaly")
            values1[i], values2[i] = np.unwrap([floater(values1[i]), floater(values2[i])])

        if "reference_eccentricity" in self.parameters:
            # Either way, we first try to make sure that the corresponding entries are floats.
            i = self.parameters.index("reference_eccentricity")
            values1[i] = metadata1.get("reference_eccentricity_bound", floaterbound(values1[i]))
            values2[i] = metadata2.get("reference_eccentricity_bound", floaterbound(values2[i]))

            if values1[i] < self.eccentricity_threshold1:
                # Then we consider metadata1 a non-eccentric system...

                # ...so we ignore the mean anomaly entirely...
                if "reference_mean_anomaly" in self.parameters:
                    i_ma = self.parameters.index("reference_mean_anomaly")
                    values1[i_ma] = values2[i_ma]

                # ...and we ignore the eccentricity if metadata2 is also non-eccentric,
                # and longer than eccentricity_threshold_penalize_shorter.
                if (
                    values2[i] < self.eccentricity_threshold2
                    and metadata2.get(
                        "number_of_orbits",
                        metadata2.get("number_of_orbits_from_start", 0)
                    ) > self.eccentricity_threshold_penalize_shorter
                ):
                    values1[i] = values2[i]

        difference = (
            np.concatenate(list(map(np.atleast_1d, values1)))
            - np.concatenate(list(map(np.atleast_1d, values2)))
        )

        if debug:
            print(f"{difference=}")

        return difference @ self.metric @ difference



# /// script
# requires-python = ">=3.12"
# dependencies = [
#     "altair==5.5.0",
#     "anywidget==0.9.15",
#     "marimo",
#     "numpy==2.1.3",
#     "pandas==2.2.3",
#     "pyarrow==19.0.1",
#     "requests==2.32.3",
#     "sxscatalog==3.0.6",
#     "traitlets==5.14.3",
# ]
# ///

import marimo

__generated_with = "0.11.19"
app = marimo.App(
    width="full",
    app_title="SXS Catalog",
    css_file="custom.css",
    html_head_file="",
)


@app.cell(hide_code=True)
def _():
    import marimo as mo

    mo.md(r"""<h1 style="text-align: center;">The SXS Catalog of Simulations</h1>""")
    return (mo,)


@app.cell(hide_code=True)
def _():
    # Import the libraries needed just for demonstrations in this notebook
    import numpy as np
    import math
    import warnings
    import pyarrow  # For efficient dataframe manipulation
    import json
    import pandas as pd
    import altair as alt
    return alt, json, math, np, pd, pyarrow, warnings


@app.cell(hide_code=True)
def _(pd):
    # Import the sxs package and load the dataframe

    # NOTE: We use the more basic `sxscatalog` package, because `sxs` cannot run in the browser
    # because it depends on `numba`, and because marimo cannot interact with the filesystem to
    # persistently cache large data files.  Normally, we would use `sxs` itself and load the
    # dataframe with `sim = sxs.load("simulations", tag={tag})`.
    import sxscatalog

    current_time = pd.Timestamp.now().strftime("%H:%M on %B %d, %Y")

    # First, we'll get the latest release so that we can write it in the notebook;
    # `sxscatalog.load("dataframe")` would do this on its own.
    latest_release = sxscatalog.simulations.Simulations.get_latest_release()
    tag_name = latest_release["tag_name"]
    release_published = pd.to_datetime(latest_release["published_at"]).strftime("%B %d, %Y")

    # Note that we load the dataframe as `df0`, so that we can filter it below as `df`
    df0 = sxscatalog.load("dataframe", tag=tag_name)

    # The dataframe is actually a sxs.SimulationsDataFrame (so that we can have attributes like df.BBH),
    # which subclasses — but is not a — pd.DataFrame, so marimo fanciness doesn't work if the dataframe
    # is just output raw.  We can get the fancy display either by calling mo.ui.dataframe(df) or by
    # acting on df with some function that returns a regular pd.DataFrame.
    return (
        current_time,
        df0,
        latest_release,
        release_published,
        sxscatalog,
        tag_name,
    )


@app.cell(hide_code=True)
def _(current_time, mo, release_published, tag_name):
    # Title and introduction
    mo.md(
        rf"""
        This page presents metadata from version {tag_name[1:]} of the SXS catalog, which was released on {release_published}, and is the current release as of <span data-tooltip="{current_time}">when you opened this page</span>.

        The metadata can be loaded into [a dataframe](https://sxs.readthedocs.io/en/main/api/simulations/#simulationsdataframe-class) using [the `sxs` package](https://github.com/sxs-collaboration/sxs/):
        ```python
        import sxs
        df = sxs.load("dataframe", tag="{tag_name}")
        ```

        That dataframe can be manipulated [as usual by pandas](https://pandas.pydata.org/pandas-docs/stable/user_guide/10min.html).
        See the [`sxs` documentation](https://sxs.readthedocs.io/en/main/tutorials/01-Simulations_and_Metadata/) for details about the metadata.

        Below, the dataframe is presented in graphical and tabular form, allowing you to explore it interactively.
        """
    ).style({"max-width": "725px", "margin": "0 auto"})
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        rf"""
        ---
        The dataframe has several useful [attributes](https://sxs.readthedocs.io/en/main/api/simulations/#simulationsdataframe-class) that allow selecting important subsets of the data.  Use the buttons below to select those subsets.
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    # Display some radio buttons that select the attributes
    system_type = mo.ui.radio(
        ["BBH", "IMR (BBH)", "BHNS", "NSNS", "any"],
        value="any",
        inline=True,
        label="System type:"
    )
    eccentricity = mo.ui.radio(
        ["eccentric", "noneccentric", "hyperbolic", "any"],
        value="any",
        inline=True,
        label="Eccentricity:"
    )
    precession = mo.ui.radio(
        ["precessing", "nonprecessing", "any"],
        value="any",
        inline=True,
        label="Precession:"
    )
    deprecation = mo.ui.radio(
        ["deprecated", "undeprecated", "any"],
        value="undeprecated",
        inline=True,
        label="Deprecation:"
    )

    df_attributes = mo.vstack([
        system_type,
        eccentricity,
        precession,
        deprecation,
    ])
    df_attributes
    return deprecation, df_attributes, eccentricity, precession, system_type


@app.cell(hide_code=True)
def _(deprecation, df0, eccentricity, precession, system_type):
    # Now, we apply the selections from above to the dataframe

    df = df0  # Keep the original around as df0

    if system_type.value == "BBH":
        df = df.BBH
    elif system_type.value == "IMR (BBH)":
        df = df.IMR
    elif system_type.value == "BHNS":
        df = df.BHNS
    elif system_type.value == "NSNS":
        df = df.NSNS

    if eccentricity.value == "eccentric":
        df = df.eccentric
    elif eccentricity.value == "noneccentric":
        df = df.noneccentric
    elif eccentricity.value == "hyperbolic":
        df = df.hyperbolic

    if precession.value == "precessing":
        df = df.precessing
    elif precession.value == "nonprecessing":
        df = df.nonprecessing

    if deprecation.value == "deprecated":
        df = type(df)(df[df["deprecated"]])
    elif deprecation.value == "undeprecated":
        df = df.undeprecated

    # And drop any columns that pandas couldn't interpret (these are the 3-vector columns)
    df = df.select_dtypes(exclude=object)

    # Finally, we re-order the columns to put these at the front
    preferred_columns = [
        "reference_mass_ratio", "reference_chi_eff", "reference_chi1_perp", "reference_chi2_perp",
        "reference_chi1_mag", "reference_chi2_mag", "reference_eccentricity", "number_of_orbits",
    ]
    columns = preferred_columns + [c for c in df.columns if c not in preferred_columns]
    df = df[columns]

    df["SXS ID"] = list(df.index)
    return columns, df, preferred_columns


@app.cell(hide_code=True)
def _(mo, simple_filtering):
    # Brief instructions for the table
    (
        mo.md("---\nClick the column headings to sort or filter by any value.")
        if simple_filtering.value
        else mo.md("---\nClick the column headings to sort.  Add a transform to filter or otherwise alter the data table.")
    )
    return


@app.cell(hide_code=True)
def _(mo):
    simple_filtering = mo.ui.checkbox(label="Simple filtering")
    simple_filtering
    return (simple_filtering,)


@app.cell(hide_code=True)
def _(df, mo, simple_filtering):
    if simple_filtering.value:
        table = mo.ui.table(df, page_size=20, show_column_summaries=True, max_columns=df.shape[1]+1)
    else:
        table = mo.ui.dataframe(df, page_size=20)
    table
    return (table,)


@app.cell(hide_code=True)
def _(mo, simple_filtering):
    mo.md(
        "We can plot selected columns from the filtered data above."
        + (
            "  If you select checkboxes in the table, only those rows will be plotted."
            if simple_filtering.value else ""
        )
    )
    return


@app.cell(hide_code=True)
def _(table):
    table_data = table.value if len(table.value) > 0 else table.data
    return (table_data,)


@app.cell(hide_code=True)
def _(df, mo):
    horizontal_axis = mo.ui.dropdown(
        options=df.columns.to_list(),
        value="reference_mass_ratio",
        label="Horizontal axis",
        allow_select_none=False,
    )
    vertical_axis = mo.ui.dropdown(
        options=df.columns.to_list(),
        value="reference_chi_eff",
        label="Vertical axis",
        allow_select_none=False,
    )
    marker_size = mo.ui.dropdown(
        options=df.columns.to_list(),
        value="reference_chi1_perp",
        label="Marker size",
    )
    marker_color = mo.ui.dropdown(
        options=df.columns.to_list(),
        value="reference_chi2_perp",
        label="Marker color",
    )

    selectors = [horizontal_axis, vertical_axis, marker_size, marker_color]

    mo.vstack([
        mo.hstack([horizontal_axis, vertical_axis], justify="space-around"),
        mo.hstack([marker_size, marker_color], justify="space-around")
    ], justify="space-around")
    return horizontal_axis, marker_color, marker_size, selectors, vertical_axis


@app.cell(hide_code=True)
def _(
    alt,
    horizontal_axis,
    marker_color,
    marker_size,
    mo,
    preferred_columns,
    selectors,
    table_data,
    vertical_axis,
):
    # Compose the plot
    used_selectors = ["SXS ID"] + preferred_columns + [
        s.value for s in selectors if s.value is not None and s.value not in preferred_columns
    ]
    #used_selectors = ["SXS ID"] + [s.value for s in selectors if s.value is not None]
    df_restricted = table_data[used_selectors]

    kwargs = dict(
        x=horizontal_axis.value,
        y=vertical_axis.value,
        tooltip=used_selectors,
    )
    if marker_size.value is not None:
        kwargs["size"] = marker_size.value
    if marker_color.value is not None:
        # Possible color schemes are listed on https://vega.github.io/vega/docs/schemes/
        kwargs["color"] = alt.Color(marker_color.value).scale(scheme="viridis")

    chart = mo.ui.altair_chart(
        alt.Chart(df_restricted, height=400)
        .mark_circle(
            stroke="#303030",
            strokeWidth=1,
            opacity=0.8,
        )
        .encode(**kwargs),
        legend_selection=True,
        label="SXS Simulations"
    )
    chart
    return chart, df_restricted, kwargs, used_selectors


@app.cell(hide_code=True)
def _(chart, mo):
    # Narrow dat the data further to just the data from (or selected in) the chart above
    chart_data = chart.value if len(chart.value) > 0 else chart.data

    (
        mo.md(
            "The following data can be further restricted by selecting a point or "
            "region on the plot; otherwise all data in the plot will be used."
        )
        if len(chart.value) == 0 else
        None
    )
    return (chart_data,)


@app.cell(hide_code=True)
def _(chart_data, math, mo, tag_name):
    # Show the code needed to load these simulations
    max_width = 6  # How many SXS IDs to allow on one line
    these_simulations =  ("this simulation" if len(chart_data)==1 else f"these {len(chart_data):,} simulations")
    load_code = (
        f"""import sxs\ndf = sxs.load("dataframe", tag="{tag_name}")\n"""
        + (
            f"""sim = sxs.load("{chart_data.index[0]}")"""
            if len(chart_data)==1 else
            (
                "sims = [sxs.load(sxs_id) for sxs_id in [\"" + "\", \"".join(chart_data.index) + "\"]]"
                if len(chart_data) < max_width else
                (
                    "sims = [sxs.load(sxs_id) for sxs_id in [\n    \""
                    + "\",\n    \"".join(
                        "\", \"".join(chart_data.index[i:min(i+max_width, len(chart_data.index))])
                        for i in range(0, math.ceil(len(chart_data.index)/max_width)*max_width, max_width)
                    )
                    + "\"\n]]"
                )
            )
        )
    )
    final_table = (
        mo.ui.table(chart_data.drop("SXS ID", axis=1), page_size=20, show_download=False)
        if len(chart_data) > 0
        else mo.md("Select a region in the plot above to see details here")
    )

    mo.vstack([
        mo.md(f"You can load {these_simulations} with\n"),
        mo.md(f"```python\n{load_code}\n```\n\n").style({"max-height": "300px", "overflow": "auto"}),
        mo.md(f"The metadata summary table for that selection follows:\n{final_table}"),
    ])
    return final_table, load_code, max_width, these_simulations


if __name__ == "__main__":
    app.run()



# SPDX-FileCopyrightText: 2025-present Mike Boyle <michael.oliver.boyle@gmail.com>
#
# SPDX-License-Identifier: MIT

from .__about__ import __version__

from . import utilities
from . import metadata
from . import simulations


def load(location, download=None, **kwargs):
    from .simulations import Simulations

    if location == "simulations":
        return Simulations.load(
            download=download,
            tag=kwargs.get("tag", ""),
            local=kwargs.get("local", False),
            annex_dir=kwargs.get("annex_dir", None),
            output_file=kwargs.get("output_file", None),
            compute_md5=kwargs.get("compute_md5", False),
            show_progress=kwargs.get("show_progress", False),
            ignore_cached=kwargs.get("ignore_cached", False),
        )

    elif location == "dataframe":
        return load("simulations", download=download, **kwargs).dataframe

    else:
        raise ValueError(
            f"\nUnknown location '{location}'. "
            "\nNote that this is `sxscatalog.load`, not `sxs.load`;"
            " perhaps you meant to use the latter."
        )



# Assemble some helper functions

def floater(x):
    import numpy as np
    try:
        f = float(x)
    except:
        f = np.nan
    return f

def floaterbound(x):
    import numpy as np
    try:
        f = float(x)
    except:
        try:
            f = float(x.replace("<", ""))
        except:
            f = np.nan
    return f

def norm(x):
    import numpy as np
    try:
        n = np.linalg.norm(x)
    except:
        n = np.nan
    return n

def three_vec(x):
    import numpy as np
    try:
        a = np.array(x, dtype=float)
        if a.shape != (3,):
            raise ValueError("Don't understand input as a three-vector")
    except:
        a = np.array([np.nan, np.nan, np.nan])
    return a

def datetime_from_string(x):
    import pandas as pd
    try:
        dt = pd.to_datetime(x)
    except:
        dt = pd.to_datetime("1970-1-1").tz_localize("UTC")
    try:
        dt = dt.tz_convert("UTC")
    except:
        pass  # No timezone information present; assuming UTC
    return dt.tz_localize(None)



"""Container interface to the catalog of SXS simulations"""

import collections
import numpy as np
import pandas as pd

from ..utilities.string_converters import *


class SimulationsDataFrame(pd.DataFrame):
    @property
    def BBH(self):
        """Restrict dataframe to just binary black hole systems"""
        return type(self)(self[self["object_types"] == "BHBH"])
    BHBH = BBH
    
    @property
    def BHNS(self):
        """Restrict dataframe to just black hole-neutron star systems"""
        return type(self)(self[self["object_types"] == "BHNS"])
    NSBH = BHNS
    
    @property
    def NSNS(self):
        """Restrict dataframe to just binary neutron star systems"""
        return type(self)(self[self["object_types"] == "NSNS"])
    BNS = NSNS

    @property
    def noneccentric(self):
        """Restrict dataframe to just non-eccentric systems (e<1e-3)"""
        return type(self)(self[self["reference_eccentricity_bound"] < 1e-3])

    @property
    def eccentric(self):
        """Restrict dataframe to just eccentric systems (e>=1e-3)"""
        return type(self)(self[self["reference_eccentricity_bound"] >= 1e-3])
    
    @property
    def nonprecessing(self):
        """Restrict dataframe to just nonprecessing systems
        
        The criterion used here is that the sum of the x-y components
        of the spins is less than 1e-3 at the reference time.
        Returns
        -------
        Simulations
            An instance of the Simulations class containing the loaded data.

        """
        return type(self)(self[
            (self["reference_chi1_perp"] + self["reference_chi2_perp"]) < 1e-3
        ])
    
    @property
    def precessing(self):
        """Restrict dataframe to just precessing systems
        
        The criterion used here is that the sum of the x-y components
        of the spins is at least 1e-3 at the reference time.
        """
        return type(self)(self[
            (self["reference_chi1_perp"] + self["reference_chi2_perp"]) >= 1e-3
        ])
    
    @property
    def IMR(self):
        """Restrict dataframe to just BBH IMR systems
        
        "IMR" stands for inspiral, merger, and ringdown.  Systems that
        will *not* be in this group include simulations that
        correspond to physical IMR systems, but were not continued
        through the merger.

        The criteria used here are just that the reference
        eccentricity and remnant mass are actual (finite) numbers.
        Currently, at least, the existence of a measured eccentricity
        means that the system is not hyperbolic or head-on.
        """
        df = self.BBH
        return type(df)(df[
            np.isfinite(df["reference_eccentricity"])
            & np.isfinite(df["remnant_mass"])
        ])
    
    @property
    def hyperbolic(self):
        """Restrict dataframe to just hyperbolic systems
        
        The criterion used here is that the (normalized) ADM mass is
        greater than 1.
        """
        total_mass = self["initial_mass1"] + self["initial_mass2"]
        normalized_ADM = self["initial_ADM_energy"] / total_mass
        return type(self)(self[
            np.isfinite(total_mass) & (total_mass > 0) & (normalized_ADM > 1)
        ])
    
    @property
    def deprecated(self):
        """Restrict dataframe to just simulations that are deprecated"""
        return type(self)(self[self["deprecated"]])
    
    @property
    def undeprecated(self):
        """Restrict dataframe to just simulations that are not deprecated"""
        return type(self)(self[~self["deprecated"]])


class Simulations(collections.OrderedDict):
    """Interface to the catalog of SXS simulations
    
    Creation
    --------
    You probably don't need to create this object yourself.  The
    easiest way to create this object is just to use the `sxs.load`
    function:

    ```python
    import sxs

    simulations = sxs.load("simulations")
    ```

    Note SXS members may also wish to read a local copy of the
    simulation annex, which can be done with
    ```python
    simulations = sxs.load("simulations", annex_dir="/path/to/SimulationAnnex.git")
    ```
    which will re-read the annex (which may take about a minute), or
    ```python
    simulations = sxs.load("simulations", local=True)
    ```
    if the annex has not been updated since the last time you
    used the `annex_dir` argument.  Once you have done this,
    calls to `sxs.load` will automatically use this local copy
    of the simulations.
    """
    releases_url = "https://api.github.com/repos/sxs-collaboration/sxscatalogdata/releases"
    url = "https://raw.githubusercontent.com/sxs-collaboration/sxscatalogdata/{tag}/simulations.json"

    def __init__(self, sims):
        """Initialize the Simulations dictionary

        Note that the constructor is not generally useful from outside
        this class.  See `Simulations.load` for a more useful
        initialization function, or simply call
        `sxs.load("simulations")`.

        """
        from ..metadata import Metadata
        super(Simulations, self).__init__(
            (k, Metadata(sims[k])) for k in sorted(sims)
        )

    @classmethod
    def get_latest_release(cls):
        """Retrieve the most-recently published release of the catalog from github"""
        import requests
        releases_response = requests.get(cls.releases_url)
        releases_response.raise_for_status()
        releases_data = releases_response.json()
        latest_release = max(releases_data, key=lambda r: r["published_at"])
        return latest_release

    @classmethod
    def local(cls, directory=None, *, download=None, output_file=None, compute_md5=False, show_progress=False):
        """Load the local catalog of SXS simulations
        
        This function loads the standard public catalog, but also
        includes any local simulations found in the given directory.
        If no directory is provided, it will look for the local
        simulations file in the sxs cache directory.

        Parameters
        ----------
        directory : {None, str, Path}, optional
            A directory containing subdirectories of SXS simulations.
            See `sxs.local_simulations` for details about what is
            expected in this directory.  If None (the default), it
            will look for the local simulations file in the sxs cache
            directory.
        download : {None, bool}, optional
            Passed to `Simulations.load` when loading the public set
            of simulations.
        output_file : {None, str, Path}, optional
            If `directory` is not None, this will be passed to
            `sxs.write_local_simulations`.
        compute_md5 : bool, optional
            If `directory` is not None, this will be passed to
            `sxs.local_simulations`.
        show_progress : bool, optional
            If `directory` is not None, this will be passed to
            `sxs.local_simulations`.

        See Also
        --------
        sxs.local_simulations : Search for local simulations
        sxs.write_local_simulations : Write local simulations to a
        file
        
        """
        import json
        from .local import write_local_simulations
        from ..utilities import sxs_directory

        if directory is not None:
            local_path = output_file
            local_simulations = write_local_simulations(
                directory,
                output_file=output_file,
                compute_md5=compute_md5,
                show_progress=show_progress
            )
        else:
            local_path = sxs_directory("cache") / "local_simulations.json"
            if not local_path.exists():
                if directory is not None:
                    raise ValueError(f"Writing local simulations for {directory=} failed")
                else:
                    raise ValueError(
                        f"Local simulations file not found, but no `directory` was provided.\n"
                        + "If called from `sxs.load`, just pass the name of the directory."
                    )
            with local_path.open("r") as f:
                local_simulations = json.load(f)
        simulations = cls.load(download)
        doi_versions = {
            k: v["DOI_versions"]
            for k,v in simulations.items()
            if "DOI_versions" in v
        }
        simulations.update(local_simulations)
        for k,v in doi_versions.items():
            simulations[k]["DOI_versions"] = v
        simulations.__file__ = str(local_path)
        return simulations

    @classmethod
    def load(
        cls,
        download=None, *,
        tag="",
        local=False,
        annex_dir=None,
        output_file=None,
        compute_md5=False,
        show_progress=False,
        ignore_cached=False,
    ):  # Make sure to pass through any parameters from `load`
        """Load the catalog of SXS simulations

        Note that — unlike most SXS data files — the simulations file
        is updated frequently.  As a result, this function — unlike
        the loading functions for most SXS data files — will download
        the simulations by default each time it is called.  However,
        also note that this function is itself cached, meaning that
        the same dict will be returned on each call in a given python
        session.  If you want to avoid that behavior, use
        `Simulations.reload`.

        Parameters
        ----------
        download : {None, bool}, optional
            If False, this function will look for the simulations in
            the sxs cache and raise an error if it is not found.  If
            True, this function will download the simulations and
            raise an error if the download fails.  If None (the
            default), it will try to download the file, warn but fall
            back to the cache if that fails, and only raise an error
            if the simulations is not found in the cache.  Note that
            this ignores the sxs configuration file entirely.

        Keyword-only Parameters
        -----------------------
        tag : str, optional
            The git tag of the simulations file to load.  If not
            provided, the latest tag will be used.  This cannot be
            provided if either `local` or `annex_dir` is provided.
        local : {None, bool}, optional
            If True, this function will load *local* simulations from
            the sxs cache.  To prepare the cache, you may wish to call
            `sxs.write_local_simulations`.
        annex_dir : {None, str, Path}, optional
            If provided and `local=True`, this function will load
            local simulations from the given directory.  This is
            equivalent to calling `Simulations.local(directory)`.
        output_file : {None, str, Path}, optional
            If `annex_dir` is not None, this will be passed to
            `sxs.write_local_simulations`.
        compute_md5 : bool, optional
            If `annex_dir` is not None, this will be passed to
            `sxs.simulations.local_simulations`.
        show_progress : bool, optional
            If `annex_dir` is not None, this will be passed to
            `sxs.simulations.local_simulations`.
        ignore_cached : bool, optional
            If True, this function will ignore the cached version of
            the simulations attached to this class, and reload the
            simulations as if the cache did not exist.

        See Also
        --------
        sxs.sxs_directory : Locate cache directory Simulations.reload

        """
        if hasattr(cls, "_simulations") and not ignore_cached:
            return cls._simulations

        import json
        import zipfile
        import warnings
        from packaging.version import Version
        from ..utilities import sxs_directory, read_config, download_file

        if tag and (local or annex_dir is not None):
            raise ValueError("Cannot specify a `tag` with `local` or `annex_dir`")

        if local or (annex_dir is not None):
            cls._simulations = cls.local(
                directory=annex_dir,
                download=download,
                output_file=output_file,
                compute_md5=compute_md5,
                show_progress=show_progress
            )
            return cls._simulations

        progress = read_config("download_progress", True)

        if download is None:
            download = True

        published_at = False
        if not tag:
            if download is not False:
                latest_release = cls.get_latest_release()
                tag = latest_release["tag_name"]
                published_at = latest_release["published_at"]
                print(
                    f"Loading SXS simulations using latest tag '{tag}', "
                    f"published at {published_at}."
                )
            else:
                tags = [
                    Version(f.stem.split("_")[-1]) # Chops off the "v" prefix so `max` works
                    for f in sxs_directory("cache").glob("simulations_*.bz2")
                ]
                if tags:
                    tag = f"v{max(tags)}"
                    warning = (
                        f"\nDownloads were turned off, so we are using the latest cached"
                        f"\nsimulations file, tagged {tag}, which may be out of date."
                    )
                    warnings.warn(warning)
                else:
                    raise ValueError(f"No simulations files found in the cache and {download=} was passed")

        # Normalize the tag to "v" followed by a normalized Version
        tag = f"v{Version(tag)}"

        cache_path = sxs_directory("cache") / f"simulations_{tag}.bz2"

        if not cache_path.exists():
            if download is False:  # Test if it literally *is* False, rather than just casts to False
                raise ValueError(f"Simulations not found in '{cache_path}' and downloading was turned off")
            else:
                # 1. Download the full json file (zipped in flight, but auto-decompressed on arrival)
                # 2. Zip to a temporary file (using bzip2, which is better than the in-flight compression)
                # 3. Replace the original simulations.zip with the temporary zip file
                # 4. Remove the full json file
                # 5. Make sure the temporary zip file is gone too
                temp_json = cache_path.with_suffix(".temp.json")
                temp_zip = cache_path.with_suffix(".temp.bz2")
                try:
                    download_file(cls.url.format(tag=tag), temp_json, progress=progress, if_newer=False)
                except Exception as e:
                    raise RuntimeError(
                        f"\nFailed to download '{cls.url.format(tag=tag)}'."
                        f"\nMaybe {tag=} does not exist?"
                    ) from e
                else:
                    if temp_json.exists():
                        with zipfile.ZipFile(temp_zip, "w", compression=zipfile.ZIP_BZIP2) as simulations_zip:
                            simulations_zip.write(temp_json, arcname="simulations.json")
                        temp_zip.replace(cache_path)
                finally:
                    temp_json.unlink(missing_ok=True)
                    temp_zip.unlink(missing_ok=True)

        if not cache_path.exists():
            raise ValueError(f"Simulations not found in '{cache_path}' for unknown reasons")

        try:
            with zipfile.ZipFile(cache_path, "r") as simulations_zip:
                try:
                    with simulations_zip.open("simulations.json") as simulations_json:
                        try:
                            simulations = json.load(simulations_json)
                        except Exception as e:
                            raise ValueError(f"Failed to parse 'simulations.json' in '{cache_path}'") from e
                except Exception as e:
                    raise ValueError(f"Failed to open 'simulations.json' in '{cache_path}'") from e
        except Exception as e:
            raise ValueError(f"Failed to open '{cache_path}' as a ZIP file") from e

        sims = cls(simulations)
        sims.__file__ = str(cache_path)
        sims.tag = tag
        if published_at:
            sims.published_at = published_at

        if not ignore_cached:
            cls._simulations = sims
        return sims

    @classmethod
    def reload(cls, *args, **kwargs):
        """Reload the catalog of SXS simulations, without caching

        Clears the cache created by a previous call to
        `Simulations.load` and returns the result of calling it again.

        All arguments are passed on to `load`.  See also the
        `ignore_cached` argument of `load`.  This does basically the
        same thing, except that it re-establishes the cache with
        whatever is reloaded.

        See Also
        --------
        Simulations.load : Caching version of this function

        """
        if hasattr(cls, "_simulations"):
            delattr(cls, "_simulations")
        return cls.load(*args, **kwargs)

    @property
    def dataframe(self):
        """Create pandas.DataFrame with metadata for all simulations

        NOTE: You might prefer to load this object directly with
        `sxs.load("dataframe")` instead of using this method.  The
        benefit to that is that `load` accepts all the same options as
        `Simulations.load`, so you can load a specific tag, local
        simulations, etc.

        The `pandas` package is the standard Python interface for
        heterogeneous data tables, like the one we have here.  This
        interface allows for more convenient slicing and querying of
        data than the list of `dict`s provided by the `Simulations`
        object.
        
        This can also be a more convenient way to access the metadata
        because the raw metadata has missing keys and mixed formats.
        Iif a key is missing from the metadata for a particular key,
        the dataframe will just have a `NaN` in that entry, rather
        than raising an exception.  Other keys may have unexpected
        entries — such as the `"reference_eccentricity"` field, which
        is *usually* a float but may be a string like "<0.0001" if the
        eccentricity is not known precisely, but is only bounded.  The
        dataframe introduces a new column called
        `"reference_eccentricity_bound"` that is always a float giving
        an upper bound on the eccentricity.

        See the `pandas` documentation for more information on how to
        use the resulting dataframe, or the `Simulations` tutorial for
        examples.
        
        """
        import numpy as np
        import pandas as pd

        if hasattr(self, "_dataframe"):
            return self._dataframe

        simulations = pd.DataFrame.from_dict(self, orient="index")

        # See also below for "number_of_orbits" field.
        # See also `sxs.metadata.metadata._backwards_compatibility`;
        # it's probably a good idea to duplicate whatever is included
        # here in that function, just to make sure nothing slips
        # through the cracks.
        for col in [
            "number_of_orbits", "number_of_orbits_from_start",
            "number_of_orbits_from_reference_time"
        ]:
            if col not in simulations.columns:
                simulations[col] = np.nan

        sims_df = SimulationsDataFrame(pd.concat((
            simulations["reference_time"].map(floater),
            simulations["reference_mass_ratio"].map(floater),
            simulations["reference_dimensionless_spin1"].map(three_vec),
            simulations["reference_dimensionless_spin1"].map(norm).rename("reference_dimensionless_spin1_mag"),
            simulations["reference_dimensionless_spin2"].map(three_vec),
            simulations["reference_dimensionless_spin2"].map(norm).rename("reference_dimensionless_spin2_mag"),
            simulations["reference_chi_eff"].map(floater),
            simulations["reference_chi1_perp"].map(floater),
            simulations["reference_chi2_perp"].map(floater),
            simulations["reference_eccentricity"].map(floater),
            simulations["reference_eccentricity"].map(floaterbound).rename("reference_eccentricity_bound"),
            simulations["reference_mean_anomaly"].map(floater),
            simulations["reference_orbital_frequency"].map(three_vec),
            simulations["reference_orbital_frequency"].map(norm).rename("reference_orbital_frequency_mag"),
            (
                simulations["reference_position1"].map(three_vec)
                -simulations["reference_position2"].map(three_vec)
            ).map(norm).rename("reference_separation"),
            simulations["reference_position1"].map(three_vec),
            simulations["reference_position2"].map(three_vec),
            simulations["reference_mass1"].map(floater),
            simulations["reference_mass2"].map(floater),
            simulations["reference_dimensionless_spin1"].map(norm).rename("reference_chi1_mag"),
            simulations["reference_dimensionless_spin2"].map(norm).rename("reference_chi2_mag"),
            simulations["relaxation_time"].map(floater),
            #simulations["merger_time"].map(floater),
            simulations["common_horizon_time"].map(floater),
            simulations["remnant_mass"].map(floater),
            simulations["remnant_dimensionless_spin"].map(three_vec),
            simulations["remnant_dimensionless_spin"].map(norm).rename("remnant_dimensionless_spin_mag"),
            simulations["remnant_velocity"].map(three_vec),
            simulations["remnant_velocity"].map(norm).rename("remnant_velocity_mag"),
            #simulations["final_time"].map(floater),
            simulations["EOS"].fillna(simulations["eos"]),
            simulations["disk_mass"].map(floater),
            simulations["ejecta_mass"].map(floater),
            simulations["object_types"].astype("category"),
            simulations["initial_data_type"].astype("category"),
            simulations["initial_separation"].map(floater),
            simulations["initial_orbital_frequency"].map(floater),
            simulations["initial_adot"].map(floater),
            simulations["initial_ADM_energy"].map(floater),
            simulations["initial_ADM_linear_momentum"].map(three_vec),
            simulations["initial_ADM_linear_momentum"].map(norm).rename("initial_ADM_linear_momentum_mag"),
            simulations["initial_ADM_angular_momentum"].map(three_vec),
            simulations["initial_ADM_angular_momentum"].map(norm).rename("initial_ADM_angular_momentum_mag"),
            simulations["initial_mass1"].map(floater),
            simulations["initial_mass2"].map(floater),
            simulations["initial_mass_ratio"].map(floater),
            simulations["initial_dimensionless_spin1"].map(three_vec),
            simulations["initial_dimensionless_spin1"].map(norm).rename("initial_dimensionless_spin1_mag"),
            simulations["initial_dimensionless_spin2"].map(three_vec),
            simulations["initial_dimensionless_spin2"].map(norm).rename("initial_dimensionless_spin2_mag"),
            simulations["initial_position1"].map(three_vec),
            simulations["initial_position2"].map(three_vec),
            #simulations["object1"].astype("category"),
            #simulations["object2"].astype("category"),
            # simulations["url"],
            #simulations["simulation_name"],
            #simulations["alternative_names"],
            # simulations["metadata_path"],
            # simulations["end_of_trajectory_time"].map(floater),
            # simulations["merger_time"].map(floater),
            simulations["number_of_orbits"].map(floater),
            simulations["number_of_orbits_from_start"].map(floater),
            simulations["number_of_orbits_from_reference_time"].map(floater),
            simulations["DOI_versions"],
            simulations["keywords"],
            simulations["date_link_earliest"].map(datetime_from_string),
            simulations["date_run_earliest"].map(datetime_from_string),
            simulations["date_run_latest"].map(datetime_from_string),
            simulations["date_postprocessing"].map(datetime_from_string),
        ), axis=1))

        # If `tag` or `published_at` are present, add them as attributes
        if hasattr(self, "tag"):
            sims_df.tag = self.tag
        if hasattr(self, "published_at"):
            sims_df.published_at = self.published_at

        # Add a column to indicate whether this simulation is deprecated
        sims_df.insert(0, "deprecated", (
            sims_df["keywords"].map(lambda ks: "deprecated" in ks)
        ))

        # See also `sxs.metadata.metadata._backwards_compatibility`;
        # it's probably a good idea to duplicate whatever is included
        # here in that function, just to make sure nothing slips
        # through the cracks.
        sims_df["number_of_orbits"] = sims_df["number_of_orbits"].fillna(
            sims_df["number_of_orbits_from_start"]
        )

        # We have ignored the following fields present in the
        # simulations.json file (as of 2024-08-04), listed here with
        # the number of non-null entries:
        #
        # alternative_names                2778
        # point_of_contact_email           2778
        # authors_emails                   2776
        # simulation_bibtex_keys           2778
        # code_bibtex_keys                 2778
        # initial_data_bibtex_keys         2778
        # quasicircular_bibtex_keys        2778
        # metadata_version                 2778
        # spec_revisions                   2778
        # spells_revision                  2778
        # merger_time                         9
        # final_time                         12
        # reference_spin1                     2
        # reference_spin2                     1
        # nitial_spin1                        2
        # initial_spin2                       2
        # remnant_spin                        2
        # initial_mass_withspin2              2
        # end_of_trajectory_time              3

        self._dataframe = sims_df
        return sims_df

    table = dataframe



from .downloads import download_file
from .sxs_directories import (
    sxs_directory, read_config, write_config
)
from .sxs_identifiers import (
    sxs_identifier_regex, sxs_identifier_re,
    lev_regex, lev_re, lev_path_re,
    sxs_id_version_lev_regex, sxs_id_version_lev_re,
    sxs_id_version_lev_exact_regex, sxs_id_version_lev_exact_re,
    sxs_path_regex, sxs_path_re,
    sxs_id, sxs_id_and_version,
    lev_number, simulation_title, sxs_id_to_url,
)


def path_to_invenio(file_path):
    """Convert a file path to an invenio-compatible name"""
    import os
    return str(file_path).replace(os.path.sep, ":")


def invenio_to_path(file_name):
    """Convert an invenio-compatible name to a file path"""
    import os
    from pathlib import Path
    return Path(file_name.replace(":", os.path.sep))
